
-- DS Heightmap (MIT license):		https://github.com/mlepage/heightmap
-- TLtools (sounds, input, resolution, A*, table serialization {ie. saving}):	https://love2d.org/wiki/TLTools
-- A* tips:							https://love2d.org/forums/viewtopic.php?f=5&t=3459&start=10#p35195
-- Spritebatch info etc:			https://love2d.org/forums/viewtopic.php?f=5&t=82562


---- TO DO ----
-- Automate ALL graphics loading
-- Map Features
	-- when applying roads, replace the terrain tile image if needed (ie road over mountain or trees looks really weird)
-- Fix the Cellular Automata NW border clumping (maybe just make it 3-5 tiles larger and skip the NW edges?)
-- Springs are too clumpy
	-- consider some kind of Voroni distribution to fix...
	-- maybe need a separate Voroni function?
-- Maps
	-- generate maps for cities, towns, ruins, caves, dungeons, etc
-- ? render "land" on minimap according to Rainfall for variety, add more grass divisions/graphics
   -- or else, just add more variety to GROUND (light grass, thick grass, flower(s), bush, etc)
-- add minimap blinking location cursor
-- UI
	-- don't let minimap cover world map
	-- reserve part of window for UI like minimap, health, stats, etc
	-- Allow minimap (and rest of UI) to be toggled on/off
-- Start using spritebatch for graphics
	-- minimap too?
---- TO DO ----


-- MAJOR GAME FUNCTIONS --

function getValue(array, x, y, z, default) -- ***
	-- *** PERFORMANCE TEST BOTH OF THESE *** --
--~ 	return x >= 1 and x <= #array and y >= 1 and y <= #array[x] and array[x][y][z] or default	-- More checks, but less table look-ups
	return array[x] and array[x][y] and array[x][y][z] or default	-- Fewer checks, but more table look-ups
	-- *** PERFORMANCE TEST BOTH OF THESE *** --
end

function changeMap()
	-- HACK *** make this flexible for ALL maps
	-- maybe define a minimap color in addition to map tiles?
	-- how to display FEATURES on minimap??

	currentMap = maps[player.map]
	collectgarbage()

	-- Load minimap colors
	local width, height = love.graphics.getDimensions()
	minimap = {x=width-currentMap.w-10, y=height-currentMap.h-10, sx=1, sy=1}
	minimap.color = {}
	minimap.color[terOcean] = {  0,   0, 255}
	minimap.color[terRiver] = { 30, 144, 255}
	minimap.color[terDesrt] = {255, 215,   0}
	minimap.color[terGrass] = {173, 255,  47}
	minimap.color[terTrees] = { 34, 139,  34}
	minimap.color[terSwamp] = {  0, 255, 255}
	minimap.color[terRocks] = {150, 150, 150}

	-- Create minimap
	-- HACK *** only display what map terrain/features have been discovered
	minimap.canvas = love.graphics.newCanvas(currentMap.w, currentMap.h)
    love.graphics.setCanvas(minimap.canvas)
        love.graphics.clear()
		for x = 1, currentMap.w do
			for y = 1, currentMap.h do
				love.graphics.setColor(minimap.color[getValue(currentMap,x,y,"terrain")])
				love.graphics.points(x, y)
			end
		end
    love.graphics.setCanvas()
end

function newGame()

	-- Initialize game creation
--~ 	love.math.setRandomSeed(10) -- HACK ***
	-- Some seeds like 1 and 3 behave VERY weirdly along the west & north borders...  WHY?? ************
	local width, height = 120, 120
	player = {map=1}

	-- Create world map
	maps = map.createNewWorld(width, height)

	-- Begin play
	changeMap()
end

function loadSpriteSheet(width, height)
	local tilesX = (width - 1) / (tileSize + 1)
	local tilesY = (height - 1) / (tileSize + 1)
	local spriteArray = {}
	for x = 1, tilesX do
		for y = 1, tilesY do
			spriteArray[(x-1) * tilesY + y] = love.graphics.newQuad((x-1)*(tileSize+1)+1, (y-1)*(tileSize+1)+1, tileSize, tileSize, width, height)
		end
	end
	return spriteArray
end


-- GUI FUNCTIONS

function playerMove(dx, dy)
	-- Kind of a hack ***
	local target = getValue(currentMap, player.x + dx, player.y + dy, "terrain")
	if not target then return end
	-- check for Creature at target location
		-- hostile creature = attack
		-- friendly creature = talk
	-- check terrain type/properties to see if movement okay
		-- if movement not okay, search tile ?
		-- if movement okay, move and check for door, etc ?
	if target > terRiver and target < terRocks or currentMap[player.x + dx][player.y + dy].feature then
		player.x = player.x + dx
		player.y = player.y + dy
	end
end

function love.load()

	-- Initialize
	ai = require "ai"
	map = require "map"
	diamondsquare = require "diamondsquare"
	love.keyboard.setKeyRepeat(true)

	-- Constants & indices
	terOcean, terRiver, terDesrt, terGrass, terTrees, terSwamp, terRocks = 1, 2, 3, 4, 5, 6, 7
--~ 	ftrRoad, ftrHmlt, ftrTown, ftrCity, ftrTunl, ftrCave, ftrRuin, ftrDngn, ftrFery = 0, 1, 2, 3, 4, 5, 6, 7, 8
	ftrRoad, ftrHmlt, ftrTown, ftrCity, ftrTunl, ftrCave, ftrRuin, ftrDngn, ftrFery = 1, 63, 64, 65, 19, 57, 55, 56, 1

	-- Initialize graphics data
	local width, height = love.graphics.getDimensions()
	tileSize = 32
	tileRadiusX, tileRadiusY = math.floor((width - tileSize)/(tileSize*2)), math.floor((height - tileSize)/(tileSize*2))
	screenPaddingX, screenPaddingY = math.floor((width - tileSize*(tileRadiusX*2+1))/2), math.floor((height - tileSize*(tileRadiusY*2+1))/2)
	-- HACK *** use the CONF file to make the window an appropriate size
	--      *** also, if the window is really big (ie. 4k monitors), apply a scalar to ALL graphics to maintain visibility

	-- Load graphics
	spriteImages = love.graphics.newImage("sprites.png")
	playerImage = love.graphics.newQuad(love.math.random(0,3)*tileSize, love.math.random(0,2)*tileSize, tileSize, tileSize, 128, 96) -- HACK *** Random sprite

	-- Load map Terrain graphics
	-- HACK *** Once this is more settled, load this in a loop and don't use pretty names here :)
	mapTerrain = {}
	mapTerrainImage = love.graphics.newImage("terrain.png")
	mapTerrain[terOcean] = love.graphics.newQuad(  1,  1, tileSize, tileSize, 133, 67)
	mapTerrain[terRiver] = love.graphics.newQuad( 34,  1, tileSize, tileSize, 133, 67)
	mapTerrain[terDesrt] = love.graphics.newQuad( 67,  1, tileSize, tileSize, 133, 67)
	mapTerrain[terGrass] = love.graphics.newQuad(100,  1, tileSize, tileSize, 133, 67)
	mapTerrain[terTrees] = love.graphics.newQuad(  1, 34, tileSize, tileSize, 133, 67)
	mapTerrain[terSwamp] = love.graphics.newQuad( 34, 34, tileSize, tileSize, 133, 67)
	mapTerrain[terRocks] = love.graphics.newQuad( 67, 34, tileSize, tileSize, 133, 67)

	-- Load map Feature graphis
	mapFeatureImage = love.graphics.newImage("feature.png")
	mapFeature = loadSpriteSheet(mapFeatureImage:getDimensions())

	-- Generate game
	newGame() -- HACK *** Waiting on UI, Save/Load, etc
end

function love.keypressed(key)
	if     key == "left"  then playerMove(-1, 0)
	elseif key == "up"    then playerMove(0, -1)
	elseif key == "right" then playerMove(1, 0)
	elseif key == "down"  then playerMove(0, 1)
	end
end

function love.draw()
	-- Reset graphics (is this still important?)
	love.graphics.setColor(255, 255, 255, 255)
	-- Draw map tiles
	for x = player.x - tileRadiusX, player.x + tileRadiusX do
		for y = player.y - tileRadiusY, player.y + tileRadiusY do
			-- Draw terrain
			local tileToDraw = getValue(currentMap, x, y, "terrain", currentMap.terDefault)
			love.graphics.draw(mapTerrainImage, mapTerrain[tileToDraw], screenPaddingX + (x-player.x+tileRadiusX)*tileSize, screenPaddingY + (y-player.y+tileRadiusY)*tileSize)
			-- Draw any map features
			tileToDraw = getValue(currentMap, x, y, "feature")
			if tileToDraw then love.graphics.draw(mapFeatureImage, mapFeature[tileToDraw], screenPaddingX + (x-player.x+tileRadiusX)*tileSize, screenPaddingY + (y-player.y+tileRadiusY)*tileSize) end
			-- Draw any creatures
			-- ****
		end
	end
	-- Draw minimap
    love.graphics.setBlendMode("alpha", "premultiplied")
    love.graphics.draw(minimap.canvas, minimap.x, minimap.y, 0, minimap.sx, minimap.sy)
    love.graphics.setBlendMode("alpha")
	-- Draw FPS
	love.graphics.print("Current FPS: "..tostring(love.timer.getFPS()), 10, 10)
	-- Draw sprites
	love.graphics.draw(spriteImages, playerImage, screenPaddingX + tileRadiusX*tileSize, screenPaddingY + tileRadiusY*tileSize)
end
